use std::{
    io::Write,
    os::fd::AsRawFd,
    sync::{Arc, Mutex},
};

use crossterm::{terminal, ExecutableCommand};

use crate::Result;

pub(crate) struct RawMode<W>
where
    W: Write + AsRawFd + Sized,
{
    write: Arc<Mutex<W>>,
}

// TODO: fork crossterm to allow using FD from as_raw_fd()
impl<W> RawMode<W>
where
    W: Write + AsRawFd + Sized,
{
    pub(crate) fn new(write_handle: Arc<Mutex<W>>) -> Result<Self> {
        terminal::enable_raw_mode()?;
        Ok(Self {
            write: write_handle,
        })
    }

    pub(crate) fn arc_mutex(self) -> Arc<Mutex<Self>> {
        Arc::new(Mutex::new(self))
    }
}

impl<W> Write for RawMode<W>
where
    W: Write + AsRawFd + Sized,
{
    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
        self.write.lock().unwrap().write(buf)
    }

    fn flush(&mut self) -> std::io::Result<()> {
        self.write.lock().unwrap().flush()
    }
}

impl<W> AsRawFd for RawMode<W>
where
    W: Write + AsRawFd + Sized,
{
    fn as_raw_fd(&self) -> std::os::fd::RawFd {
        self.write.lock().unwrap().as_raw_fd()
    }
}

/*
impl<W> ExecutableCommand for RawMode<W>
where
    W: Write + AsRawFd + Sized,
{
    fn execute(&mut self, command: impl crossterm::Command) -> std::io::Result<&mut Self> {
        let mut write = self.write.lock().unwrap();
        match write.execute(command) {
            Ok(_) => {
                drop(write);
                Ok(self)
            }
            Err(e) => Err(e),
        }
    }
}
*/

impl<W> Drop for RawMode<W>
where
    W: Write + AsRawFd + Sized,
{
    fn drop(&mut self) {
        terminal::disable_raw_mode().unwrap();
    }
}
